/*
 * Copyright (c) 2001, 2002, 2003
 *    Digital Stratum.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Company, Digital Stratum, nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * PGLOGD v2.3
 *
 * Log web server entries to a PostgreSQL database.
 *
 * Add this line to your httpd.conf:
 *
 *	LogFormat "%t %T %>s %b %m %v %h \"%U\" \"%{Referer}i\" \"%{User-agent}i\" \"%r\" %l %u" pglogd
 *
 * Then anywhere you need a log:
 *
 *	CustomLog "/path/to/apache/logs/pglogd_fifo" pglogd
 *
 *
 * This file uses 132 columns, sorry to the 80 column people...
 *
 *
 * TABLE FORMAT
 *
 * Active (no indexes)
 *
 * log_entries(
 * request_time timestamp,
 * seconds_to_serve int4,
 * response_code int4,
 * bytes_sent int4,
 * request_type varchar(10),
 * virtual_domain varchar(80),
 * remote_host varchar(80),
 * request_url_path varchar(80),
 * referer text,
 * user_agent text,
 * request text,
 * logname varchar(30),
 * username varchar(30));
 *
 *
 * Stats (full info plus indexes)
 *
 * request_time timestamp,
 * seconds_to_serve int4,
 * response_code int4,
 * bytes_sent int4,
 * request_type varchar(10),
 * virtual_domain varchar(80),
 * remote_host varchar(80),
 * host_type varchar(3),
 * request_url_path text,
 * referer text,
 * referer_domain varchar(80),
 * referer_sengine varchar(80),
 * user_agent text,
 * request text,
 * logname varchar(30),
 * username varchar(30));
 */
#ifdef SOLARIS
char * strsep(char **stringp, const char *delim);
#include <strings.h>
#else
#include <string.h>
#endif

#include <libpq-fe.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include "pglogd.h"


/*
 * Global state variables.
 */
int iSignal;	/* Did we receive a signal. */
int iDaemon;	/* Are we a daemon. */


/*
 * Miscellaneous globals.
 */
int iMaxfiles;	/* Maximum files open by a uid, if not specified. */
char cEntry[BUFFSIZE];
char *cpAppName;

/*
 * The user's configuration.
 */
struct user_config sUserConfig;


/*
 * Main program.
 */
int
main(int argc, char **argv)
{
	/******
	 *
	 * BEGIN DATA.
	 *
	 ******/

	/*
	 * Local daemon state variables.
	 */
	pid_t		pPid;			/* Pid for the daemon. */


	/*
	 * Ascynronous management and select() variables.
	 */
	struct timeval tvPoll;		/* timeval structure to poll select(). */
	fd_set setRead;				/* File descriptors ready to read. */
	fd_set setWrite;			/* File descriptors ready to write. */
	fd_set setExcp;				/* File descriptors with socket out-of-band data. */
	int iMAXfd;					/* Largest open file descriptor. */
	int iDescriptorsReady;		/* Number of file descriptors ready for processing. */


	/*
	 * Database connection management.
	 */
	PGconn *pgConn;				/* Database Connection. */
	PGresult *pgRes;			/* Handle to a query result set. */
	int fdPGsock;				/* Handle to the db connection socket. */
	time_t iBackendDown;		/* Set to the time the backend went down, otherwise 0 if the backend is up. */
	int iBackendBusy;			/* Is the backend busy with a previous query. */
	PostgresPollingStatusType dbPollFlag;
	int iDBresetting;			/* TRUE if the backend connection in a state of being reset. */
	int iDBread;				/* Include the backend connection for reading. */
	int iDBwrite;				/* Include the backend connection for writing. */


	/*
	 * FIFO file and data-flow management.
	 */
	int fdFIFO;					/* FIFO file descriptor. */
	int fdFIFO_OBSD_WRITE;		/* FIFO file descriptor for writing (OpenBSD only). */
	struct stat sFIFOstatus;	/* FIFO status buffer. */
	char cReadBuffer[BUFFSIZE];	/* The read() buffer. */
	char *cpReadOffset;			/* Location in the read buffer to begin store data on next read(). */
	char *cpEOL;				/* End Of Line of a log entry in the read buffer. */
	ssize_t stBytesRead;		/* Number of bytes read from the FIFO during the last read(). */
	int	iBytesInBuffer;			/* Number of bytes already in the read buffer. */
	int iEatData;				/* Ignore FIFO data because an entry was not found within the buffer size. */
	int iHaveReadData;			/* TRUE if there is still read buffer data to process. */

	/*
	 * Overflow file management.
	 */
	int iOverflowRead;			/* Overflow file old entries. */
	int iOverflowWrite;			/* Overflow file new entries. */
	char cOverflowBuffer[BUFFSIZE];	/* The overflow read() buffer. */
	ssize_t stOverflowRead;		/* Number of bytes read from the overflow file. */
	off_t tReadStart;			/* Starting position in the overflow file prior to a read. */
	off_t tReadOffset;			/* Current read position in the overflow file. */
	int iOverflowLock;			/* Prevents overflow entry processing from doubling up. */
	int iOverflowEntry;			/* Indicates the current entry in an overflow entry. */
	int iOverFlowQueued;		/* Log entries written to the overflow file, i.e. waiting to be written. */
	int iOverFlowProcessed;		/* Log entries read from the overflow file and written to the database. */


	/*
	 * Log entry management.
	 */
	int iHaveEntry;				/* TRUE if there a log entry ready to parse. */
	char *cpQuery;				/* Query buffer. */
	struct pglog_entry sLogEntry;


	/*
	 * Command line options.
	 */
	char *cpConfigFile;			/* The user supplied configuration file. */
	int iSkipDatabase;			/* Whether or not to skip the database connection at start-up. */


	/*
	 * General purpose.
	 */
	int i;						/* Miscellaenous loop counter. */
	int iRetVal;				/* Shutdown return value. */


	/******
	 *
	 * END DATA.
	 *
	 ******/


	/*
	 * Initialize our state, non-daemon.
	 */
	iSignal = 0;
	iDaemon = 0;

	/*
	 * One time local variable initialization.
	 */
	cpAppName = &(*argv[0]);

	iMAXfd = 0;
	fdFIFO = -1;
	fdFIFO_OBSD_WRITE = -1;
	iOverflowRead = -1;
	iOverflowWrite = -1;

	pgConn = NULL;
	dbPollFlag = PGRES_POLLING_WRITING;
	iBackendDown = 0;
	iBackendBusy = FALSE;
	iDBresetting = FALSE;
	iDBread = FALSE;
	iDBwrite = FALSE;

	cpReadOffset = cReadBuffer;
	iBytesInBuffer = 0;
	iEatData = FALSE;
	iHaveReadData = FALSE;

	iOverflowLock = FALSE;
	iOverflowEntry = FALSE;
	iOverFlowQueued = 0;
	iOverFlowProcessed = 0;

	iHaveEntry = FALSE;
	cpQuery = NULL;

	cpConfigFile = NULL;
	iSkipDatabase = FALSE;


	/*
	 * Check for our two options:
	 *
	 *  1. Skip connecting to the database at statup.
	 *  2. Read from a specified config file.
	 */
	for ( i = 1 ; i < argc ; i++ )
	{
		if ( strncmp(&(*argv[i]), "-c", 2) == 0 )
		{
			cpConfigFile = &(*argv[i + 1]);
			i++;
		}

		else if ( strncmp(&(*argv[i]), "-s", 2) == 0 )
			iSkipDatabase = TRUE;
	}


	/*
	 * Get the initial configuration parameters.
	 */
	read_config(cpConfigFile);


	/*
	 * Make sure the FIFO file does not exist.
	 */
	if ( stat(sUserConfig.cFIFO, &sFIFOstatus) < 0 )
	{
		/*
		 * Make sure the error is "file does not exist"
		 */
		if ( errno != ENOENT )
		{
			fprintf(stderr, "%s: Fatal stat() error [%d]\n", cpAppName, errno);
			exit(ERR_FIFOSTAT);
		}
	}

	else
	{
		/*
		 * If the FIFO exists, the server may already be running, so
		 * inform the user and exit.
		 */
		fprintf(stderr, "%s: FIFO Named Pipe [%s] already exists, server already running?\n", cpAppName, sUserConfig.cFIFO);
		exit(ERR_FIFOEXISTS);
	}


	/*
	 * Make sure we can open and write to the overflow file.
	 */
	if ( (iRetVal = open_overflow(&iOverflowRead, &iOverflowWrite, 1)) != ERR_NORMAL )
		exit(iRetVal);


	/*
	 * Check that we can connect to the database.
	 */
	if ( iSkipDatabase == FALSE )
	{
		pgConn = PQconnectdb(sUserConfig.cCONNECTION);
		if ( (iRetVal = checkConnStatus(pgConn, sUserConfig.cDBNAME, cpAppName)) != ERR_NORMAL )
		{
			PQfinish(pgConn);
			exit(iRetVal);
		}
		PQfinish(pgConn);
	}


	/*
	 * Make sure we can aquire the FIFO file.
	 */
	umask(0);
	if ( mkfifo(sUserConfig.cFIFO, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH) < 0 )
	{
		fprintf(stderr, "%s: FIFO Named Pipe [%s] could not be created.  mkfifo() error [%d]\n", cpAppName, sUserConfig.cFIFO, errno);
		exit(ERR_FIFOCREATE);
	}


	/*
	 * Become a deamon, muhaaaa
	 */
	if ( (pPid = fork()) < 0 )
	{
		fprintf(stderr, "%s: Fatal error, could not fork() [%d]\n", cpAppName, errno);
		exit(ERR_FORK);
	}

	else if ( pPid != 0 )
		exit(ERR_NORMAL);


	/*
	 * Parent is gone, the child becomes the session leader.
	 */
	setsid();
	chdir("/");
	umask(S_IWGRP|S_IWOTH);
	iDaemon = 1;

	logentry("Resuming normal operation.", 0);


	/*
	 * Close up all open files.
	 */
	if ( (iMaxfiles = sysconf(_SC_OPEN_MAX)) < 0 )
		iMaxfiles = MAXFILES;

	for ( i = 0 ; i < iMaxfiles ; i++ )
	{
		if ( close(i) < 0 )
		{
			if ( errno == EINTR )
				continue;
			else
				break;
		}
	}

	logentry("Closed inherited file descriptiors:", i);


	/*
	 * Establish our signal handlers.
	 */
	iSignal = 0;

	if ( signal(SIGHUP, sig_init) == SIG_ERR )
	{
		logentry("Can't catch SIGHUP", errno);
		iRetVal = ERR_CATCH_SIGHUP;
		goto done;
	}

	if ( signal(SIGINT, sig_init) == SIG_ERR )
	{
		logentry("Can't catch SIGINT", errno);
		iRetVal = ERR_CATCH_SIGINT;
		goto done;
	}

	if ( signal(SIGTERM, sig_init) == SIG_ERR )
	{
		logentry("Can't catch SIGTERM", errno);
		iRetVal = ERR_CATCH_SIGTERM;
		goto done;
	}

	/*
	 * Don't let the system restart our select() or read() call, we need to
	 * do this so we can shutdown properly.
	 */
	siginterrupt(SIGHUP, 1);
	siginterrupt(SIGINT, 1);
	siginterrupt(SIGTERM, 1);


	/*
	 * Connect to the database.
	 */
	if ( iSkipDatabase == FALSE )
	{
		pgConn = PQconnectdb(sUserConfig.cCONNECTION);
		if ( (iRetVal = checkConnStatus(pgConn, sUserConfig.cDBNAME, cpAppName)) != ERR_NORMAL )
			goto done;
		snprintf(cEntry, BUFFSIZE, "Database connection [%s] opened.", sUserConfig.cDBNAME);
		logentry(cEntry, 0);

		/*
		 * Get the database socket to use with select().
		 */
		if ( (fdPGsock = PQsocket(pgConn)) < 0 )
		{
			logentry("PQsocket() error.", -1);
			iRetVal = ERR_DBCONN;
			goto done;
		}
	}

	else
	{
		snprintf(cEntry, BUFFSIZE, "Skipping initial database connection [%s].", sUserConfig.cCONNECTION);
		logentry(cEntry, 0);

		iDBresetting = TRUE;
		iDBread = FALSE;
		iDBwrite = FALSE;

		pgConn = NULL;
		fdPGsock = -1;

		time(&iBackendDown);
	}



	/*
	 * Set the database connection to non-blocking.
	 *
	 * Tom Lane says this is broken, so for now our PQsendQuery()
	 * calls will block while the data is being sent.
	 */
	/*
	if ( PQsetnonblocking(pgConn, TRUE) == -1 )
	{
		logentry("PQsetnonblocking() error.", -1);
		PQfinish(pgConn);
		exit(ERR_DBNBLOCK);
	}
	 */


	/*
	 * Open up the FIFO file.  Open for read and write since an End Of File
	 * would be received if there are no process with the FIFO open for
	 * writing.
	 */
	#ifndef OPENBSD
	if ( (fdFIFO = open(sUserConfig.cFIFO, O_RDWR)) < 0 )
	{
		snprintf(cEntry, BUFFSIZE, "FIFO Named Pipe [%s] could not be opened for read/write.  open() error:\n", sUserConfig.cFIFO);
		logentry(cEntry, errno);
		iRetVal = ERR_FIFOOPEN;
		goto done;
	}
	#else
	// Seems OpenBSD can't open a FIFO as O_RDWR, so we'll do it as
	// two discrete file descriptors.
	if ( (fdFIFO = open(sUserConfig.cFIFO, O_RDONLY)) < 0 )
	{
		snprintf(cEntry, BUFFSIZE, "FIFO Named Pipe [%s] could not be opened for reading.  open() error:\n", sUserConfig.cFIFO);
		logentry(cEntry, errno);
		iRetVal = ERR_FIFOOPEN;
		goto done;
	}

	if ( (fdFIFO_OBSD_WRITE = open(sUserConfig.cFIFO, O_WRONLY)) < 0 )
	{
		snprintf(cEntry, BUFFSIZE, "FIFO Named Pipe [%s] could not be opened for writing.  open() error:\n", sUserConfig.cFIFO);
		logentry(cEntry, errno);
		iRetVal = ERR_FIFOOPEN;
		goto done;
	}
	#endif

	snprintf(cEntry, BUFFSIZE, "FIFO Named Pipe [%s] opened.", sUserConfig.cFIFO);
	logentry(cEntry, fdFIFO);


	/*
	 * Open the overflow file, it should already exist.
	 */
	if ( (iRetVal = open_overflow(&iOverflowRead, &iOverflowWrite, 2)) != ERR_NORMAL )
		goto done;
	logentry("Overflow file open for reading:", iOverflowRead);
	logentry("Overflow file open for writing:", iOverflowWrite);

	tReadStart = tReadOffset = lseek(iOverflowRead, 0, SEEK_CUR);

	if ( (iOverFlowQueued = init_overflow(iOverflowRead)) < ERR_NORMAL )
		goto done;
	logentry("Overflow file entries found queued:", iOverFlowQueued);


	/*
	 * Sit on our butt and wait for the webserver client to send
	 * us a log entry, or process overflow entries.
	 */

	while ( iSignal == 0 )
	{
		/*
		 * Reset the file descriptor sets.  We are watching the FIFO
		 * and the backend's socket.
		 */
		FD_ZERO(&setRead);
		FD_ZERO(&setWrite);
		FD_ZERO(&setExcp);

		FD_SET(fdFIFO, &setRead);

		/*
		 * Never submit both reading and writing at the same time, and
		 * never run a select() on a bad socket (database backend down.)
		 */
		if ( iBackendDown == 0 )
		{
			if ( iDBwrite == FALSE && iDBread == TRUE )
				FD_SET(fdPGsock, &setRead);

			if ( iDBwrite == TRUE && iDBread == FALSE )
				FD_SET(fdPGsock, &setWrite);

			FD_SET(fdPGsock, &setExcp);
		}

		/*
		 * Get the highest numbered descriptor assigned.  Do this each
		 * time since a re-opened db connection might change descriptors.
		 */
		if ( fdFIFO > fdPGsock || iBackendDown != 0 )
			iMAXfd = fdFIFO + 1;
		else
			iMAXfd = fdPGsock + 1;

		/*
		 * If there are no overflow entries, the select can block.  BUT, make
		 * sure we wake up at least every MAXSELECTDELAY seconds in case the
		 * backend goes down.  Also, if we still have read data to process
		 * then don't block.
		 */
		if ( (iOverFlowQueued > iOverFlowProcessed && iOverflowEntry == FALSE
			&& iBackendBusy == FALSE && iDBresetting == FALSE ) || iHaveReadData == TRUE )
		{
			/*
			 * For polling with select(), pass a zero value time stucture.
			 */
			tvPoll.tv_sec = 0;
			/* tvPoll.tv_usec = 100000; testing */
			tvPoll.tv_usec = 0;
		}

		else
		{
			tvPoll.tv_sec = sUserConfig.iMAXSELECTDELAY;
			tvPoll.tv_usec = 0;
		}

		if ( (iDescriptorsReady = select(iMAXfd, &setRead, &setWrite, &setExcp, &tvPoll)) < 0 )
		{
			if ( errno != EINTR )
			{
				/*
				 * Unrecoverable select error.
				 */
				logentry("select() error:", errno);
				iRetVal = ERR_SELECT;
				break;  /* goto done. */
			}

			else if ( iSignal == 0 )
				continue;
			else
				break;	/* goto done. */
		}


		/*
		 * Select() says one or more of our descriptors is ready, or
		 * we did a poll-type select() because we have overflow entries
		 * or read-data to process.
		 */
		if ( iDescriptorsReady == 0 && iHaveReadData == FALSE && iBackendBusy == FALSE
			&& iDBresetting == FALSE && iOverflowEntry == FALSE && iOverFlowQueued > iOverFlowProcessed )
		{
			/*
			 * Everything seems quiet, so let's make some noise! :)
			 * Feed an overflow entry into the mix.
			 */

			/*
			 * Remember where we started so we can:
			 *	1. Reset if necessary.
			 *	2. Seek back to the end of the entry if we read to much data.
			 */
			tReadStart = tReadOffset;

			restart_overflow:

				if ( (stOverflowRead = read(iOverflowRead, cOverflowBuffer, (size_t)BUFFSIZE)) <= 0 )
				{
					/*
					 * Catch EINTR and go back to reading the Overflow file.
					 */
					if ( stOverflowRead == -1 )
					{
						if ( errno != EINTR )
						{
							logentry("Overflow file read() error:", errno);
							iRetVal = ERR_OVERFLOWREAD;
							break;	/* goto done. */
						}

						else if ( iSignal == 0 )
							goto restart_overflow;
						else
						{
							iRetVal = ERR_NORMAL;
							break;	/* goto done. */
						}
					}

					else
					{
						/*
						 * All out of overflow entries!
						 */
						logentry("Error, overflow file [End Of File] received before all entries processed:",
							iOverFlowProcessed);
						logentry("Resetting overflow file, number of ignored entries written:",
							iOverFlowQueued - iOverFlowProcessed);

						close(iOverflowRead);
						close(iOverflowWrite);

						if ( unlink(sUserConfig.cOVERFLOWFILE) < 0 )
						{
							logentry("Overflow file unlink() error:", errno);
							break;	/* goto done. */
						}

						if ( (iRetVal = open_overflow(&iOverflowRead, &iOverflowWrite, 3)) != ERR_NORMAL )
							break;	/* goto done. */
						tReadOffset = lseek(iOverflowRead, 0, SEEK_CUR);

						iOverFlowQueued = 0;
						iOverFlowProcessed = 0;

						continue;
					}
				}

			// restart_overflow

			/*
			 * The man page says that read() will give us our whole buffer
			 * if reading a normal file and that many bytes are left in the
			 * file.  So, since only valid entries are written to the overflow
			 * file and our read buffer is the same as the maximum allowed
			 * entry size, there should be a full entry in here somewhere.
			 * Find the newline, and reset the read position in the file.
			 */
			if ( (cpEOL = memchr(cOverflowBuffer, LF, stOverflowRead)) != NULL )
			{
				/*
				 * Chew up the end of line character and set the file
				 * offset position.
				 */
				*(cpEOL) = '\0';

				if ( (tReadOffset = lseek(iOverflowRead, tReadStart + (cpEOL - cOverflowBuffer) + 1, SEEK_SET)) < 0 )
				{
					logentry("Overflow file lseek() error:", errno);
					iRetVal = ERR_OVERFLOWREAD;
					break;	/* goto done. */
				}

				/*
				 * Parse the entry.
				 */
				if ( (iRetVal = parse_entry(cOverflowBuffer, &sLogEntry)) == TRUE )
				{
					iOverflowEntry = TRUE;
					iHaveEntry = TRUE;
				}

				else if (iRetVal == FALSE )
				{
					free_entry(&sLogEntry);
					continue;
				}

				else
					break;	/* memory error, goto done. */
			}

			else
			{
				/*
				 * This is bad, we can't process the overflow file, so shut down
				 * to get someone's attention.
				 */
				logentry("Overflow file no entry found in buffer.", 0);
				iRetVal = ERR_OVERFLOWREAD;
				break;	/* goto done. */
			}
		}


		/*
		 * Check for all overflow entries processed.
		 */
		if ( iOverFlowQueued != 0 && iOverFlowQueued == iOverFlowProcessed )
		{
			/*
			 * All overflow entries processed, check for EOF on the read.
			 */
			restart_eof_check:

				if ( (stOverflowRead = read(iOverflowRead, cOverflowBuffer, (size_t)BUFFSIZE)) < 0 )
				{
					/*
					 * Catch EINTR and go back to reading the Overflow file.
					 */
					if ( stOverflowRead == -1 )
					{
						if ( errno != EINTR )
						{
							logentry("Overflow file read() error:", errno);
							iRetVal = ERR_OVERFLOWREAD;
							break;	/* goto done. */
						}

						else if ( iSignal == 0 )
							goto restart_eof_check;
						else
							break;	/* goto done. */
					}
				}

			// restart_eof_check

			if ( stOverflowRead != 0 )
				logentry("Error, overflow file [End Of File] not received, but all entries processed.  Bytes remaining:",
					stOverflowRead);

			logentry("Resetting overflow file, all written entries processed:", iOverFlowProcessed);

			close(iOverflowRead);
			close(iOverflowWrite);

			if ( unlink(sUserConfig.cOVERFLOWFILE) < 0 )
			{
				logentry("Overflow file unlink() error:", errno);
				break;	/* goto done. */
			}

			if ( (iRetVal = open_overflow(&iOverflowRead, &iOverflowWrite, 3)) != ERR_NORMAL )
				break;	/* goto done. */
			tReadOffset = lseek(iOverflowRead, 0, SEEK_CUR);

			iOverFlowQueued = 0;
			iOverFlowProcessed = 0;
		}


		/*
		 * Check the FIFO.
		 */
		if ( FD_ISSET(fdFIFO, &setRead) != 0 || iHaveReadData == TRUE )
		{
			if ( iHaveReadData == FALSE )
			{
				/*
				 * select() says the FIFO is ready to read so read() should not block.
				 */
				restart_fifo:

					if ( (stBytesRead = read(fdFIFO, cpReadOffset, (size_t)(BUFFSIZE - iBytesInBuffer - 1))) <= 0 )
					{
						/*
						 * Catch EINTR and go back to reading the FIFO or shut down.
						 * Any other error exits.
						 */
						if ( stBytesRead == -1 )
						{
							if ( errno != EINTR )
							{
								logentry("FIFO read() error:", errno);
								iRetVal = ERR_FIFOREAD;
								break;	/* goto done. */
							}

							else if ( iSignal == 0 )
								goto restart_fifo;
							else
								break;	/* goto done. */
						}

						else
						{
							/*
							 * This should not happen, but if it does...
							 */
							logentry("FIFO End Of File received.", 0);
							break;	/* goto done. */
						}
					}

				// restart_fifo

				cpReadOffset += stBytesRead;
				iBytesInBuffer = (cpReadOffset - cReadBuffer);
			}

			/*
			 * Find a newline.  If none exists, we do not have an entry yet,
			 * unless we are at the buffer limit, in which case we just eat
			 * the input until we find a newline, then start over.
			 */
			if ( (cpEOL = memchr(cReadBuffer, LF, iBytesInBuffer)) != NULL )
			{
				iHaveReadData = TRUE;

				if ( iEatData == FALSE )
				{
					/*
					 * Chew up the end of line characters.
					 */
					if ( *(cpEOL - 1) == CR )
						*(cpEOL - 1) = '\0';
					else
						*(cpEOL) = '\0';

					/*
					 * Parse the entry to make sure it is indeed an entry.
					 */
					if ( (iRetVal = parse_entry(cReadBuffer, &sLogEntry)) == TRUE )
						iHaveEntry = TRUE;
					else if ( iRetVal == FALSE )
						free_entry(&sLogEntry);
					else
						break;	/* memory error, goto done. */
				}

				/*
				 * Retain any data that may have been read in following the EOL we just found.
				 */
				if ( (cpEOL + 1) < cpReadOffset )
				{
					iBytesInBuffer = (cpReadOffset - (cpEOL + 1));
					bcopy((cpEOL + 1), cReadBuffer, iBytesInBuffer);
					cpReadOffset = (cReadBuffer + iBytesInBuffer);
				}

				else
				{
					iBytesInBuffer = 0;
					cpReadOffset = cReadBuffer;
				}

				/*
				 * Once we find an EOL we can start looking for an entry again.
				 */
				iEatData = FALSE;
			}

			else
				iHaveReadData = FALSE;

			if ( iBytesInBuffer >= (BUFFSIZE - 1) )
			{
				iBytesInBuffer = 0;
				cpReadOffset = cReadBuffer;
				iEatData = TRUE;
			}
		}


		/*
		 * Perform any connection restart actions.  Keep the server
		 * from trying a reconnect every time there is server activity,
		 * wait at least 10 seconds between attempts.
		 */
		if ( iDBresetting == TRUE && time(NULL) >= (iBackendDown + 10) )
		{
			/*
			 * Processing an attempt to restart a backend connection.
			 */
			poll_again:

				dbPollFlag = PQconnectPoll(pgConn);

				switch (dbPollFlag)
				{
					case PGRES_POLLING_ACTIVE:
						goto poll_again;
						break;

					case PGRES_POLLING_OK:
						/*
						 * Connection is up!
						 */
						iBackendDown = 0;
						iBackendBusy = FALSE;
						iDBresetting = FALSE;
						iDBread = FALSE;
						iDBwrite = FALSE;

						if ( iOverflowEntry == TRUE )
						{
							/*
							 * Looks like we were trying to write an overflow
							 * entry when the back-end went down.  So reset
							 * the read() so we can try again.
							 */
							iOverflowEntry = FALSE;

							if ( (tReadOffset = lseek(iOverflowRead, tReadStart, SEEK_SET)) < 0 )
							{
								logentry("Overflow file lseek() error:", errno);
								iRetVal = ERR_OVERFLOWREAD;
								goto done;
							}
						}

						/*
						 * Get the database socket to use with select().
						 */
						if ( (fdPGsock = PQsocket(pgConn)) < 0 )
						{
							logentry("PQsocket() error.", -1);
							time(&iBackendDown);
							iDBresetting = FALSE;
							iDBread = FALSE;
							iDBwrite = FALSE;
						}

						else
							logentry("Backend connection up.", 0);

						break;

					case PGRES_POLLING_READING:
						iDBread = TRUE;
						iDBwrite = FALSE;

						break;

					case PGRES_POLLING_WRITING:
						iDBread = FALSE;
						iDBwrite = TRUE;

						break;

					default:
						/*
						 * The restart failed, so on the next pass try
						 * to reconnect again.
						 */
						iDBresetting = FALSE;
						iDBread = FALSE;
						iDBwrite = FALSE;

						/*
						 * Reset the time so the next attempt will be
						 * delayed at least 10 seconds.
						 */
						time(&iBackendDown);

						logentry("Backend reconnect failed.", dbPollFlag);

						break;
				}

			// poll_again:
		}

		/*
		 * If the backend is down or just went down, attempt a restart.
		 */
		else if ( iBackendDown != 0 || PQstatus(pgConn) == CONNECTION_BAD )
		{
			/*
			 * Keep attempting a reconnect until the backend comes back up.
			 */
			if ( iBackendDown == 0 )
			{
				time(&iBackendDown);
				logentry("Backend connection down.", 0);
			}

			if ( pgConn != NULL )
			{
				fdPGsock = -1;
				PQfinish(pgConn);
				pgConn = NULL;
			}

			/*
			 * Keep the server from calling PQconnectStart repeatedly until
			 * it has had a chance to attempt PQconnectPoll above.
			 */
			if ( dbPollFlag != PGRES_POLLING_WRITING )
			{
				if ( (pgConn = PQconnectStart(sUserConfig.cCONNECTION)) == NULL )
					logentry("PQconnectStart() failed.", 0);
				else
				{
					/*
					 * Make sure we know we need to poll the next time through.
					 */
					dbPollFlag = PGRES_POLLING_WRITING;
					iDBread = FALSE;
					iDBwrite = TRUE;
					iDBresetting = TRUE;
				}
			}
		}

		/*
		 * Check if waiting on a previous query result.
		 */
		else if ( FD_ISSET(fdPGsock, &setRead) != 0 )
		{
			if ( PQconsumeInput(pgConn) == 0 )
			{
				logentry("PQconsumeInput() error.", 0);
				logentry("Backend connection down.", 0);
				time(&iBackendDown);
			}

			else
			{
				while ( (iBackendBusy = PQisBusy(pgConn)) == 0 )
				{
					/*
					 * Process all results sent from the backend.
					 */
					if ( (pgRes = PQgetResult(pgConn)) != NULL )
					{
						checkResStatus(pgConn, pgRes, cpAppName, cpQuery);
						PQclear(pgRes);
					}

					else
					{
						/*
						 * Done reading the socket until a new query is submitted.
						 */
						iDBread = FALSE;
						iDBwrite = FALSE;

						if ( iOverflowEntry == TRUE )
						{
							iOverflowEntry = FALSE;
							iOverFlowProcessed++;

							if ( iOverFlowProcessed % sUserConfig.iOVERFLOW_WARNING == 0 )
								logentry("NOTICE, Overflow file entries processed: ", iOverFlowProcessed);
						}

						break;
					}
				}
			}
		}


		/*
		 * Check to see if we have a log entry to process.
		 */
		if ( iHaveEntry == TRUE )
		{
			/*
			 * See where to write the entry.
			 */
			entry_retry:

				if ( iBackendDown == 0 && iBackendBusy == FALSE )
				{
					/*
					 * Write to the database.
					 */

					i = QUERYBASE;
					i += (strlen(sLogEntry.request_url_path) * 2);
					i += (strlen(sLogEntry.referer) * 2);
					i += (strlen(sLogEntry.user_agent) * 2);
					i += (strlen(sLogEntry.request) * 2);

					/*
					 * Leave the query in cpQuery until we need to make a new one.
					 * This is so we can log the query if there is a problem.
					 */
					if ( cpQuery != NULL )
						free(cpQuery);

					if ( (cpQuery = malloc((size_t) (sizeof(char) * i))) == NULL )
					{
						logentry("malloc() error while building query:", errno);
						iRetVal = ERR_MEMORY;
						break;	/* goto done. */
					}

					/*
					 * Build up the query.
					 */
					i = 0;
					i += sprintf(cpQuery, "insert into log_entries values ('");
					i += escapeSpecial(sLogEntry.request_time, &cpQuery[i]);
					i += sprintf(&cpQuery[i], "',%d,%d,%d,'",
						sLogEntry.seconds_to_serve, sLogEntry.response_code, sLogEntry.bytes_sent);

					i += escapeSpecial(sLogEntry.request_type, &cpQuery[i]);
					i += sprintf(&cpQuery[i], "','");

					i += escapeSpecial(sLogEntry.virtual_domain, &cpQuery[i]);
					i += sprintf(&cpQuery[i], "','");

					i += escapeSpecial(sLogEntry.remote_host, &cpQuery[i]);
					i += sprintf(&cpQuery[i], "','");

					i += escapeSpecial(sLogEntry.request_url_path, &cpQuery[i]);
					i += sprintf(&cpQuery[i], "','");

					i += escapeSpecial(sLogEntry.referer, &cpQuery[i]);
					i += sprintf(&cpQuery[i], "','");

					i += escapeSpecial(sLogEntry.user_agent, &cpQuery[i]);
					i += sprintf(&cpQuery[i], "','");

					i += escapeSpecial(sLogEntry.request, &cpQuery[i]);
					i += sprintf(&cpQuery[i], "','");

					i += escapeSpecial(sLogEntry.logname, &cpQuery[i]);
					i += sprintf(&cpQuery[i], "','");

					i += escapeSpecial(sLogEntry.username, &cpQuery[i]);
					i += sprintf(&cpQuery[i], "')");


					/*
					 * Store in the database.  Blocks while dispatching query.
					 */
					if ( PQsendQuery(pgConn, cpQuery) == FALSE )
					{
						logentry("PQsendQuery() error.", FALSE);
						logentry("Backend connection down.", 0);
						time(&iBackendDown);
						goto entry_retry;
					}

					/* PQnonblocking() is broken, so we don't need this right now.
					if ( (iRetVal = PQflush(pgConn)) != 0 )
					{
						logentry("PQflush() error:", iRetVal);
						time(&iBackendDown);
						goto entry_retry;
					}
					 */

					/*
					 * Make sure we don't attempt to write to the database
					 * again until this query is complete.
					 */
					iBackendBusy = TRUE;

					/*
					 * Set our socket to check when the query is finished.
					 */
					iDBread = TRUE;
					iDBwrite = FALSE;
				}

				else if ( iOverflowEntry == FALSE )
				{
					/*
					 * Write to the overflow file.
					 */
					i = strlen(sLogEntry.original);

					if ( write(iOverflowWrite, sLogEntry.original, i) != i )
					{
						logentry("Failure writing to overflow file.", errno);
						iRetVal = ERR_OVERFLOWWRITE;
						break;	/* goto done. */
					}

					else if ( write(iOverflowWrite, "\n", 1) != 1 )
					{
						logentry("Failure writing to overflow file.", errno);
						iRetVal = ERR_OVERFLOWWRITE;
						break;	/* goto done. */
					}

					iOverFlowQueued++;

					if ( iOverFlowQueued % sUserConfig.iOVERFLOW_WARNING == 0 )
						logentry("WARNING, Overflow file entries:", iOverFlowQueued);
				}

				else
				{
					/*
					 * Can't write the overflow entry to the database, so back up
					 * the read() so we can try again.
					 */
					iOverflowEntry = FALSE;

					if ( (tReadOffset = lseek(iOverflowRead, tReadStart, SEEK_SET)) < 0 )
					{
						logentry("Overflow file lseek() error:", errno);
						iRetVal = ERR_OVERFLOWREAD;
						break;	/* goto done. */
					}
				}

			// entry_retry:

			free_entry(&sLogEntry);
			iHaveEntry = FALSE;
		}

	} // While iSignal == 0 (no signals.)

	done: // Multi-level escape hatch for fatal errors.

	/*
	 * Close the overflow file, the FIFO, and the database connection.
	 */
	if ( iOverflowRead >= 0 )
		close(iOverflowRead);

	if ( iOverflowWrite >= 0 )
		close(iOverflowWrite);

	if ( fdFIFO >= 0 )
	{
		close(fdFIFO);
		unlink(sUserConfig.cFIFO);
	}

	PQfinish(pgConn);
	free_entry(&sLogEntry);

	if ( iRetVal == ERR_NORMAL )
		logentry("Normal Shutdown", 0);
	else
		logentry("Abnormal Shutdown", iRetVal);

	logentry("Server State: errno:", errno);
	logentry("Server State: iHaveEntry:", iHaveEntry);
	logentry("Server State: iHaveReadData:", iHaveReadData);
	logentry("Server State: iOverflowEntry:", iOverflowEntry);
	logentry("Server State: iOverFlowQueued:", iOverFlowQueued);
	logentry("Server State: iOverFlowProcessed:", iOverFlowProcessed);
	logentry("Server State: iBackendDown:", iBackendDown);
	logentry("Server State: iBackendBusy:", iBackendBusy);


	return(iRetVal);
}


/*
 * Checks the status of a database connect.
 */
int
checkConnStatus(PGconn *pgConn, const char *dbName, char *cpProg)
{
	if ( PQstatus(pgConn) == CONNECTION_BAD )
	{
		if ( iDaemon == 0 )
		{
			fprintf(stderr, "%s: Connection to database <%s> failed.\n", cpProg, dbName);
			fprintf(stderr, "%s: %s", cpProg, PQerrorMessage(pgConn));
		}

		else
		{
			snprintf(cEntry, BUFFSIZE,
				"%s: Connection to database <%s> failed.\n%s: %s", cpProg, dbName, cpProg, PQerrorMessage(pgConn));
			logentry(cEntry, 0);
		}

		return(ERR_DBCONN);
	}

	return(ERR_NORMAL);
}


/*
 * Checks the status of a query.
 */
int
checkResStatus(PGconn *pgConn, PGresult *pgRes, char *cpProg, char *cpQuery)
{
	if ( !pgRes || ((PQresultStatus(pgRes) != PGRES_EMPTY_QUERY) &&
					(PQresultStatus(pgRes) != PGRES_COMMAND_OK) &&
					(PQresultStatus(pgRes) != PGRES_TUPLES_OK)) )
	{
		if ( iDaemon == 0 )
		{
			fprintf(stderr, "%s: Query Failed\n", cpProg);
			fprintf(stderr, "%s: Offending Query - <%s>\n", cpProg, cpQuery);
			fprintf(stderr, "%s: %s", cpProg, PQresultErrorMessage(pgRes));

			return(ERR_DBQUERY);
		}

		else
		{
			snprintf(cEntry, BUFFSIZE,
				"%s: Query Failed\n%s: Offending Query - <%s>\n%s: %s",
				cpProg, cpProg, cpQuery, cpProg, PQerrorMessage(pgConn));
			logentry(cEntry, 0);

			return(ERR_DBQUERY);
		}

	}

	return(ERR_NORMAL);
}


/*
 * Escapes the following: " ' \
 *
 * The output buffer is a terminated string.  Its length is returned.
 */
int
escapeSpecial(char *cpIn, char *cpOut)
{
	int i;
	int iCount;
	int iOffset;	/* Keep track of our position in the output buffer. */

	iCount = strlen(cpIn);
	iOffset = 0;

	for ( i = 0 ; i < iCount ; i++ )
	{
		if ( cpIn[i] == '"' || cpIn[i] == '\'' || cpIn[i] == '\\' )
			cpOut[iOffset++] = '\\';

		cpOut[iOffset++] = cpIn[i];
	}

	cpOut[iOffset] = '\0';

	return(iOffset);
}


/*
 * Create and/or open the overflow file.
 *
 * iMode:
 *	1 - Create only
 *	2 - Open only
 *	3 - Create and open
 */
int
open_overflow(int *iRead, int *iWrite, int iMode)
{
	if ( iMode != 2 )
	{
		/*
		 * Make sure we can open and write to the overflow file.
		 */
		if ( ((*iWrite) = open(sUserConfig.cOVERFLOWFILE,
			O_WRONLY | O_APPEND | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) < 0 )
		{
			if ( iDaemon == 0 )
				fprintf(stderr, "%s: Overflow file [%s] could not be opened or created.  open() error [%d]\n",
					cpAppName, sUserConfig.cOVERFLOWFILE, errno);
			else
			{
				snprintf(cEntry, BUFFSIZE, "Overflow file [%s] could not be opened or created.  open() error:",
					sUserConfig.cOVERFLOWFILE);
				logentry(cEntry, errno);
			}

			return(ERR_OVERFLOW);
		}
		close((*iWrite));
	}


	if ( iMode != 1 )
	{
		/*
		 * Open the overflow file, it should already exist.  We will open it
		 * twice, once for reading and once for writing (so we can read and
		 * write to the file at the same time without having to moving the
		 * offset.)
		 */
		if ( ((*iRead) = open(sUserConfig.cOVERFLOWFILE, O_RDONLY)) < 0 )
		{
			snprintf(cEntry, BUFFSIZE, "Overflow file [%s] could not be opened for reading.  open() error:\n",
				sUserConfig.cOVERFLOWFILE);
			logentry(cEntry, errno);

			return(ERR_OVERFLOW);
		}

		if ( ((*iWrite) = open(sUserConfig.cOVERFLOWFILE, O_WRONLY | O_APPEND)) < 0 )
		{
			snprintf(cEntry, BUFFSIZE, "Overflow file [%s] could not be opened for writing.  open() error:\n",
				sUserConfig.cOVERFLOWFILE);
			logentry(cEntry, errno);

			return(ERR_OVERFLOW);
		}
	}

	return(ERR_NORMAL);
}


/*
 * Returns the number of entries in the file.  This should be used
 * to preset the number of entries queued.  The return value should
 * be >= ERR_NORMAL otherwise and error occured.
 */
int
init_overflow(int iRead)
{
	size_t stBytesRead;
	int iCount;
	char cBuffer[BUFFSIZE];
	char *cpOffset;

	iCount = 0;

	/*
	 * Make sure to start at the beginning of the file.
	 */
	if ( lseek(iRead, 0, SEEK_SET) < 0 )
	{
		logentry("Overflow file lseek() error:", errno);
		return(ERR_OVERFLOWREAD);
	}


	while ( iCount >= 0 )
	{
		if ( (stBytesRead = read(iRead, cBuffer, (size_t)BUFFSIZE)) < 0 )
		{
			/*
			 * Catch EINTR and go back to reading the Overflow file.
			 */
			if ( stBytesRead == -1 )
			{
				if ( errno != EINTR )
				{
					logentry("Overflow file initialization read() error:", errno);
					return(ERR_OVERFLOWREAD);
				}

				else if ( iSignal == 0 )
					continue;
				else
					return(ERR_OVERFLOWREAD);
			}
		}

		if ( stBytesRead == 0 )
		{
			/*
			 * At the EOF, reset the read pointer so we can process these entries.
			 */
			if ( lseek(iRead, 0, SEEK_SET) < 0 )
			{
				logentry("Overflow file lseek() error:", errno);
				return(ERR_OVERFLOWREAD);
			}

			return(iCount);
		}

		cpOffset = cBuffer;

		while ( (cpOffset = memchr(cpOffset, LF, (stBytesRead - (cpOffset - cBuffer)))) != NULL )
		{
			cpOffset++;
			iCount++;
		}
	}

	logentry("Error, Overflow file contains more than 2,147,483,648 entries.", 0);
	return(ERR_OVERFLOWREAD);
}


/*
 * Parse out a pglogd specific log entry.  Deviation from CLF
 * ("Common Log Format") was chosen because CLF puts user entered
 * data prior to machine generated data, i.e. the date, host, etc.
 * Thus, CLF makes it unnecessarily difficult to parse a log entry
 * because a malitious user could enter a username similar to this:
 *
 * 10.0.81.14 - - [07/Jul/2001
 *
 * Which is perfectally legal, however would most likely stop *all*
 * log parsing programs dead in their tracks.  For more details, see
 * the README and pglogd's home page.
 *
 * pglogd format:
 *
 * %t %T %>s %b %m %v %h \"%U\" \"%{Referer}i\" \"%{User-agent}i\" \"%r\" %l %u
 *
 * pglog structure:
 *	struct pglog_entry {
 *		char *original;
 *		char request_time[31];
 *		int seconds_to_serve;
 *		int response_code;
 *		int bytes_sent;
 *		char request_type[11];
 *		char virtual_domain[81];
 *		char remote_host[81];
 *		char *request_url_path;
 *		char *referer;
 *		char *user_agent;
 *		char *request;
 *		char logname[31];
 *		char username[31];
 *	};
 *
 * Active table structure:
 *
 * log_entries(
 * request_time timestamp,
 * seconds_to_serve int4,
 * response_code int4,
 * bytes_sent int4,
 * request_type varchar(10),
 * virtual_domain varchar(80),
 * remote_host varchar(80),
 * request_url_path varchar(80),
 * referer text,
 * user_agent text,
 * request text,
 * logname varchar(30),
 * username varchar(30));
 *
 *
 * Returns true if the entry could be parsed, otherwise false.  If there
 * was a critical error, ERR_MEMORY will be returned.
 */
int
parse_entry(char *cpReadBuffer, struct pglog_entry *spLogEntry)
{
	char *cpBuffer;
	char *cpHead;
	char *cpTail;
	char *cpTemp;
	int iDots;

	cpBuffer = NULL;
	spLogEntry->original = NULL;
	spLogEntry->request_url_path = NULL;
	spLogEntry->referer = NULL;
	spLogEntry->user_agent = NULL;
	spLogEntry->request = NULL;

	/*
	 * Copy the buffer so we can tear it up and retain the original.
	 */
	if ( (cpBuffer = malloc((size_t) (sizeof(char) * (strlen(cpReadBuffer) + 1)))) == NULL )
		goto memory_error;

	if ( (spLogEntry->original = malloc((size_t) (sizeof(char) * (strlen(cpReadBuffer) + 1)))) == NULL )
		goto memory_error;

	bcopy(cpReadBuffer, cpBuffer, strlen(cpReadBuffer) + 1);
	bcopy(cpReadBuffer, spLogEntry->original, strlen(cpReadBuffer) + 1);

	/*
	 * Ahh, machine generated data...
	 */
	cpHead = cpBuffer;

	/*
	 * Entry date stamp.
	 */
	if ( *cpHead++ != '[' )
		goto fail;

	if ( (cpTail = memchr(cpHead, ']', 31)) == NULL )
		goto fail;

	*cpTail = '\0';
	strncpy(spLogEntry->request_time, cpHead, (cpTail - cpHead) + 1);


	/*
	 * Seconds to respond.  If we don't have a response time
	 * in 6 digits then, whao!, something is wrong!
	 *
	 * Skip the bracket ']' and space delimiter.
	 */
	cpHead = cpTail + 2;

	if ( (cpTail = memchr(cpHead, ' ', 7)) == NULL )
		goto fail;

	*cpTail = '\0';
	spLogEntry->seconds_to_serve = atoi(cpHead);


	/*
	 * Response code.  Codes, so far, are only 3 digits.
	 *
	 * Skip the space delimiter.
	 */
	cpHead = cpTail + 1;

	if ( (cpTail = memchr(cpHead, ' ', 4)) == NULL )
		goto fail;

	*cpTail = '\0';
	spLogEntry->response_code = atoi(cpHead);


	/*
	 * Bytes sent.  This *could* be rather large, so let's
	 * say that 4 Gig in a single request would cover just
	 * about every possible situation.  If not, I would like
	 * to meet the person who can suck down 4+ Gig!
	 *
	 * Skip the space delimiter.
	 */
	cpHead = cpTail + 1;

	if ( (cpTail = memchr(cpHead, ' ', 11)) == NULL )
		goto fail;

	*cpTail = '\0';
	spLogEntry->bytes_sent = atoi(cpHead);


	/*
	 * Request type.  Methods specified in the HTTP/1.1 RFC:
	 * "OPTIONS", "GET", "POST", "PUT", "DELETE", "TRACE", "CONNECT"
	 *
	 * Skip the space delimiter.
	 */
	cpHead = cpTail + 1;

	if ( (cpTail = memchr(cpHead, ' ', 11)) == NULL )
		goto fail;

	*cpTail = '\0';
	strncpy(spLogEntry->request_type, cpHead, (cpTail - cpHead) + 1);


	/*
	 * Virtual domain.
	 *
	 * Skip the space delimiter.
	 */
	cpHead = cpTail + 1;

	if ( (cpTail = memchr(cpHead, ' ', 81)) == NULL )
		goto fail;

	*cpTail = '\0';

	/*
	 * If we are only recording the TLD, step back two '.'
	 */
	if ( sUserConfig.iTLDONLY == TRUE )
	{
		cpTemp = cpTail;
		iDots = 0;

		/*
		 * Note the use of short circuit here to keep cpTemp on the
		 * second dot.
		 */
		while ( iDots < 2 && --cpTemp >= cpHead )
		{
			if ( *cpTemp == '.' )
				iDots++;
		}

		cpHead = cpTemp + 1;
	}

	strncpy(spLogEntry->virtual_domain, cpHead, (cpTail - cpHead) + 1);


	/*
	 * Remote Host.  This *may* be a full host name even if HostNameLookups
	 * is off.  Check Apache's web site for details, but the jist is
	 * that if any hostname auth is configured, then a hostname lookup
	 * is performed regardless of what HostNameLookups is set to...
	 *
	 * Skip the space delimiter.
	 */
	cpHead = cpTail + 1;

	if ( (cpTail = memchr(cpHead, ' ', 81)) == NULL )
		goto fail;

	*cpTail = '\0';
	strncpy(spLogEntry->remote_host, cpHead, (cpTail - cpHead) + 1);


	/*
	 * Requested URL path.  Now we start to get into user-land data... :(
	 *
	 * Skip the space delimiter.
	 */
	cpHead = cpTail + 1;

	if ( *cpHead++ != '"' )
		goto fail;

	if ( (cpTail = strstr(cpHead, "\" ")) == NULL )
		goto fail;

	*cpTail = '\0';

	if ( (spLogEntry->request_url_path = malloc((size_t) (sizeof(char) * ((cpTail - cpHead) + 1)))) == NULL )
		goto memory_error;

	strncpy(spLogEntry->request_url_path, cpHead, (cpTail - cpHead) + 1);


	/*
	 * Referer.
	 *
	 * Skip the previous closing quote ('"') and space delimiter.
	 */
	cpHead = cpTail + 2;

	if ( *cpHead++ != '"' )
		goto fail;

	if ( (cpTail = strstr(cpHead, "\" ")) == NULL )
		goto fail;

	*cpTail = '\0';

	if ( (spLogEntry->referer = malloc((size_t) (sizeof(char) * ((cpTail - cpHead) + 1)))) == NULL )
		goto memory_error;

	strncpy(spLogEntry->referer, cpHead, (cpTail - cpHead) + 1);


	/*
	 * User agent.
	 *
	 * Skip the previous closing quote ('"') and space delimiter.
	 */
	cpHead = cpTail + 2;

	if ( *cpHead++ != '"' )
		goto fail;

	if ( (cpTail = strstr(cpHead, "\" ")) == NULL )
		goto fail;

	*cpTail = '\0';

	if ( (spLogEntry->user_agent = malloc((size_t) (sizeof(char) * ((cpTail - cpHead) + 1)))) == NULL )
		goto memory_error;

	strncpy(spLogEntry->user_agent, cpHead, (cpTail - cpHead) + 1);


	/*
	 * Request.  Ahh, the request, haven for user data.
	 *
	 * Skip the previous closing quote ('"') and space delimiter.
	 */
	cpHead = cpTail + 2;

	if ( *cpHead++ != '"' )
		goto fail;

	if ( (cpTail = strstr(cpHead, "\" ")) == NULL )
		goto fail;

	*cpTail = '\0';

	if ( (spLogEntry->request = malloc((size_t) (sizeof(char) * ((cpTail - cpHead) + 1)))) == NULL )
		goto memory_error;

	strncpy(spLogEntry->request, cpHead, (cpTail - cpHead) + 1);


	/*
	 * Logname.  From an ident server if available.  No spaces allowed
	 * here, at least *we* will only consider characters up to the first
	 * space or 30 characters, what ever comes first.
	 *
	 * Skip the previous closing quote ('"') and space delimiter.
	 */
	cpHead = cpTail + 2;

	if ( (cpTail = memchr(cpHead, ' ', 31)) == NULL )
		goto fail;

	*cpTail = '\0';
	strncpy(spLogEntry->logname, cpHead, (cpTail - cpHead) + 1);


	/*
	 * Username.  The last great haven for user data.  Read the rest of
	 * the line, up to 30 characters, the rest is destine for /dev/null.
	 *
	 * Skip the space delimiter.
	 */
	cpHead = cpTail + 1;

	strncpy(spLogEntry->username, cpHead, 30);
	spLogEntry->username[30] = '\0';

	free(cpBuffer);
	return(TRUE);

	fail:
		free(cpBuffer);
		return(FALSE);

	memory_error:

		if ( cpBuffer != NULL )
			free(cpBuffer);

		free_entry(spLogEntry);
		logentry("System malloc() error:", errno);
		return(ERR_MEMORY);
}


/*
 * Releases a pglog_entry structure's dynamic memory.
 */
void
free_entry(struct pglog_entry *spLogEntry)
{
	if ( spLogEntry->original != NULL )
		free(spLogEntry->original);

	if ( spLogEntry->request_url_path != NULL )
		free(spLogEntry->request_url_path);

	if ( spLogEntry->referer != NULL )
		free(spLogEntry->referer);

	if ( spLogEntry->user_agent != NULL )
		free(spLogEntry->user_agent);

	if ( spLogEntry->request != NULL )
		free(spLogEntry->request);

	return;
}


/*
 * Let the main loop know we were signaled.
 */
static void
sig_init(int iSig)
{
	logentry("Signal caught:", iSig);
	iSignal = 1;
	return;
}


/*
 * Write to the error log.
 */
void
logentry(const char *cpEntry, int iData)
{
	FILE *fpLog;				/* Daemon log. */
	char cTime[30];				/* Time stamp. */
	time_t tTime;

	tTime = time(NULL);
	#ifdef SOLARIS
	ctime_r(&tTime, cTime, 30);
	#else
	ctime_r(&tTime, cTime);
	#endif
	cTime[strlen(cTime) - 1] = '\0';

	fpLog = fopen(sUserConfig.cLOGFILE, "a");
	fprintf(fpLog, "[%s] %s [%d]\n", cTime, cpEntry, iData);
	fclose(fpLog);

	return;
}


/*
 * Read the user configuration file.
 *
 *	struct user_config {
 *		char cDBNAME[256];
 *		char cCONNECTION[256];
 *		char cFIFO[256];
 *		char cLOGFILE[256];
 *		char cOVERFLOWFILE[256];
 *		int iOVERFLOW_WARNING;
 *		int iMAXSELECTDELAY;
 *		int iTLDONLY;
 *	};
 *
 * The rules:
 *  Skip all white space except in the last (3rd) parameter.
 *  Skip any line that does not begin with a keyword.
 *  Skip any line longer that 256 characters.
 *
 * To skip a config file and just use the defaults, pass
 * a NULL as the file name.
 */
void
read_config(const char *cpFile)
{
	FILE *fpConfig;
	int iEatData;
	int i;

	char cBuffer[256];
	char *cpBuffer;
	char *cpTokens[3];
	char **cpPointer;

	/*
	 * Set the defaults in case the parsing fails.
	 */
	strncpy(sUserConfig.cDBNAME, DBNAME, 255);
	sUserConfig.cDBNAME[255] = '\0';

	strncpy(sUserConfig.cCONNECTION, CONNECTION, 255);
	sUserConfig.cCONNECTION[255] = '\0';

	strncpy(sUserConfig.cFIFO, FIFO, 255);
	sUserConfig.cFIFO[255] = '\0';

	strncpy(sUserConfig.cLOGFILE, LOGFILE, 255);
	sUserConfig.cLOGFILE[255] = '\0';

	strncpy(sUserConfig.cOVERFLOWFILE, OVERFLOWFILE, 255);
	sUserConfig.cOVERFLOWFILE[255] = '\0';

	sUserConfig.iOVERFLOW_WARNING = OVERFLOW_WARNING;
	sUserConfig.iMAXSELECTDELAY = MAXSELECTDELAY;
	sUserConfig.iTLDONLY = TLDONLY;

	/*
	 * If the file name is NULL, just use the defaults.
	 */
	if ( cpFile == NULL )
		return;

	/*
	 * Try to open the config file given.
	 */
	if ( (fpConfig = fopen(cpFile, "r")) == NULL )
	{
		if ( iDaemon == 0 )
			fprintf(stderr, "%s: Configuration file [%s] could not be opened for reading.  Using defaults.\n",
				cpAppName, cpFile);

		else
		{
			snprintf(cEntry, BUFFSIZE,
				"Configuration file [%s] could not be opened for reading.  Using defaults.", cpFile);
			logentry(cEntry, 0);
		}

		return;
	}


	/*
	 * Read each line and look for a known keyword.
	 */
	iEatData = FALSE;

	while ( fgets(cBuffer, 256, fpConfig) != NULL )
	{
		/*
		 * Determine if we have a whole line and skip any lines longer than
		 * our maximum line, currently hard-coded to 256 characters.
		 */
		if ( cBuffer[strlen(cBuffer) - 1] != '\n' )
		{
			if ( !feof(fpConfig) || iEatData == TRUE )
			{
				iEatData = TRUE;
				continue;
			}
		}

		else if ( iEatData == TRUE )
		{
			iEatData = FALSE;
			continue;
		}

		else
			cBuffer[strlen(cBuffer) - 1] = '\0';

		/*
		 * We have a line within our maxline parameters, so tokenize
		 * a keyword, the assignment '=', and the parameter.
		 */
		cpBuffer = cBuffer;

		cpTokens[0] = NULL;
		cpTokens[1] = NULL;
		cpTokens[2] = NULL;
		cpPointer = cpTokens;

		while ( (*cpPointer = strsep(&cpBuffer, " \t")) != NULL )
		{
			if ( **cpPointer != '\0' )
			{
				if ( ++cpPointer >= &cpTokens[2] )
					break;
			}
		}

		/*
		 * Did we run out of tokens?  If not, clean any white space off of
		 * the last parameter.
		 */
		if ( cpBuffer != NULL )
			cpTokens[2] = (cpBuffer + strspn(cpBuffer, " \t"));

		/*
		 * Skip blank lines, lines with too few tokens, and
		 * lines that don't start with an alphabetic character.
		 */
		if ( cpTokens[0] == NULL || cpTokens[1] == NULL || cpTokens[2] == NULL )
			continue;

		else
		{
			if ( *(cpTokens[1]) != '=' )
				continue;

			strtolower(cpTokens[0]);

			if ( *(cpTokens[0]) < 97 || *(cpTokens[0]) > 122 )
				continue;
		}


		/*
		 * Find a keyword.
		 */
		for ( i = 0 ; i < NUM_KEYWORDS ; i++ )
		{
			if ( strncmp(cKeywords[i], cpTokens[0], strlen(cKeywords[i])) == 0 )
				break;
		}

		switch ( i )
		{
			case dbname:

				strncpy(sUserConfig.cDBNAME, cpTokens[2], 255);
				sUserConfig.cDBNAME[255] = '\0';

				break;

			case connection:

				strncpy(sUserConfig.cCONNECTION, cpTokens[2], 255);
				sUserConfig.cCONNECTION[255] = '\0';

				break;

			case fifo:

				strncpy(sUserConfig.cFIFO, cpTokens[2], 255);
				sUserConfig.cFIFO[255] = '\0';

				break;

			case logfile:

				strncpy(sUserConfig.cLOGFILE, cpTokens[2], 255);
				sUserConfig.cLOGFILE[255] = '\0';

				break;

			case overflowfile:

				strncpy(sUserConfig.cOVERFLOWFILE, cpTokens[2], 255);
				sUserConfig.cOVERFLOWFILE[255] = '\0';

				break;

			case overflow_warning:

				if ( atoi(cpTokens[2]) >= 10 )
					sUserConfig.iOVERFLOW_WARNING = atoi(cpTokens[2]);

				break;

			case maxselectdelay:

				if ( atoi(cpTokens[2]) >= 10 )
					sUserConfig.iMAXSELECTDELAY = atoi(cpTokens[2]);

				break;

			case tldonly:

				sUserConfig.iTLDONLY = true_or_false(cpTokens[2]);

				break;
		} // End switch for keyword

	} // End while config file lines to read

	fclose(fpConfig);

	return;
}


/*
 * FUNCTION:  strtolower()
 *  Make a string all lowercase.
 *
 * CALLED WITH:
 *  cpString    - Pointer to a null terminated string.
 *
 * MODIFIES:
 *  cpString    - Pointer to lowercase string.
 *
 * RETURNS:
 *  0
 *  1 if cpSting is NULL.
 *
 * NOTES:
 *  This function modifies the original string.
 */
int
strtolower(char *cpString)
{
    int iX = 0;
    int iMax = 0;

    if ( cpString == NULL )
        return(1);

    if ( (iMax = strlen(cpString)) == 0 )
        return(0);

    for (iX = 0 ; iX < iMax ; iX++)
        *(cpString + iX) = tolower(*(cpString + iX));

    return(0);
}


/*
 * Determins if a string is true or fales by checking
 * for several values:
 *
 * Yes, No, True, False, 0, or 1.  Case does not matter.
 */
int
true_or_false(const char *cpString)
{
	char cBuffer[10];

	strncpy(cBuffer, cpString, 10);
	cBuffer[9] = '\0';

	strtolower(cBuffer);

	if ( strncmp(cBuffer, "yes", 3) == 0 )
		return(TRUE);

	if ( strncmp(cBuffer, "no", 2) == 0 )
		return(FALSE);

	if ( strncmp(cBuffer, "true", 4) == 0 )
		return(TRUE);

	if ( strncmp(cBuffer, "false", 5) == 0 )
		return(FALSE);

	if ( strncmp(cBuffer, "0", 1) == 0 )
		return(FALSE);

	return(TRUE);
}


/*
 * Local STRSEP from the FreeBSD libc for those OSes that don't have
 * their own STRSEP, namely SOLARIS.
 */

/*
 * Copyright (c) 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Get next token from string *stringp, where tokens are possibly-empty
 * strings separated by characters from delim.
 *
 * Writes NULs into the string at *stringp to end tokens.
 * delim need not remain constant from call to call.
 * On return, *stringp points past the last NUL written (if there might
 * be further tokens), or is NULL (if there are definitely no more tokens).
 *
 * If *stringp is NULL, strsep returns NULL.
 */
#ifdef SOLARIS

char *
strsep(stringp, delim)
	register char **stringp;
	register const char *delim;
{
	register char *s;
	register const char *spanp;
	register int c, sc;
	char *tok;

	if ((s = *stringp) == NULL)
		return (NULL);
	for (tok = s;;) {
		c = *s++;
		spanp = delim;
		do {
			if ((sc = *spanp++) == c) {
				if (c == 0)
					s = NULL;
				else
					s[-1] = 0;
				*stringp = s;
				return (tok);
			}
		} while (sc != 0);
	}
	/* NOTREACHED */
}
#endif
